:- module( pubmed, [ pubmed_cited_by/2, pubmed_cited_by/3,
                     pubmed_cites/2, pubmed_cites/3,
                     pubmed_summary_info/3,
                     pubmed_search/2, pubmed_search/3,
                     pubmed_summary_display/1, pubmed_summary_display/2, pubmed_summary_display/3,
                     pubmed_version/2
                   ]
         ).

/**  <module> connect to publication services at pubmed

   A simple library for communicating with pubmed publications. 
   Currently allows (a) searching on conjunctions and disjunctions, (b) fetching the details of a pubmed id(s)
   (c) the publications citing a paper and (d) simple reporting of multiple fetched information.
   It requires the curl executable to be in the path. Only tested on Linux. 
   It is being developed on SWI-Prolog 6.1.8 but should also work on Yap Prolog.

     @version 0.0.3, 2012/08/15
     @author Nicos Angelopoulos
     @see http://bioinformatics.nki.nl/~nicos/sware/pubmed
     @see http://www.ncbi.nlm.nih.gov/books/NBK25500/
     @see examples.pl file in the packs directory
     @see sources at http://bioinformatics.nki.nl/~nicos/sware/pubmed/pubmed-0.0.3.tgz
*/

:- ensure_loaded( library(sgml) ).

% Section: defaults, shortcuts.

eutils( 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/' ).

default_names( Names ) :- 
     Names = ['Author','Title','Source','Pages','PubDate',
              'Volume','Issue','ISSN','PmcRefCount',
              'PubType','FullJournalName'].

pubmed_search_defaults( [verbose(false),retmax(100),tmp_keep(false)] ).

pubmed_summary_display_defaults( [display(['Title','Author']),names(Names)] ) :-
     default_names( Names ).
     
% Section: interface predicates

/** pubmed_version( +Version, +Date ).

Get version information and date of publication.
*/
pubmed_version( 0:0:3, date(2012,08,15) ).

/** pubmed_search( +STerm, -Ids ).

Short form of pubmed_search( +STerm, -Ids, [] ).

*/
pubmed_search( STerm, Ids ) :-
     pubmed_search( STerm, Ids, [] ).

/** pubmed_search( +STerm, -Ids, Options ).

     Search in pubmed for terms in the search term =STerm=.
     In this, conjunction is marked by _|, (comma)|_ and 
     disjunction by _|;|_ (semi-column). '-' pair terms are considered as 
     Key-Value and interpreted as Value[Key] in the query.
     List are thought to be flat conjoint search terms  with no pair values in them which are 
     interpreted by pubmed also as OR operations.
     (See example below.)
     Known keys are : =journal=, =pdat=. =au=, =|All Fields|=
     The predicate constructs a query that is posted via the http API provided
     by NCBI (http://www.ncbi.nlm.nih.gov/books/NBK25500/).
     =Options= can be a single term or list of terms from : 

     * retmax(RetMax)  the maximum number of records that will be returned def: 100

     * verbose(Verbose) if =|Verbose == true|= then the predicate verbose
                         about its progress by, for instance, requesting 
                         query is printed on current output stream.

     * tmp_file(Tmp)  file to use, or when =Tmp= is variable the file that was used
                      to receive the results from pubmed.
     * tmp_keep(Keep) keep the file with the xml result iff =|Keep==true|=

     * qtranslation(QTrans) return in =QTrans= the actual query ran on the 
                      the pubmed server.
     For instance, taking an example from the url we show how to find 
     all breast cancer articles that were published in Science in 2008.

==
?-
    St = (journal=science,[breast,cancer],pdat=2008),
    pubmed_search( St, Ids, [verbose(true),qtranslation(QTrans)] ),
    length( Ids, Len ), write( number_of:Len ), nl.

http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=pubmed&retmax=100&term=science\[journal\]+AND+breast+cancer+AND+2008\[pdat\]
process_create(path(curl),[-o,/tmp/pl_13858_1,http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=pubmed&retmax=100&term=science\[journal\]+AND+breast+cancer+AND+2008\[pdat\]],[])
  % Total    % Received % Xferd  Average Speed   Time    Time     Time  Current
                                 Dload  Upload   Total   Spent    Left  Speed
100  3008    0  3008    0     0   3585      0 --:--:-- --:--:-- --:--:--  4641
tmp_file(/tmp/pl_13858_1)
number_of:6
St = (journal=science, [breast, cancer], pdat=2008),
Ids = ['19008416', '18927361', '18787170', '18487186', '18239126', '18239125'],
QTrans = ['("Science"[Journal] OR "Science (80- )"[Journal] OR "J Zhejiang Univ Sci"[Journal]) AND ("breast neoplasms"[MeSH Terms] OR ("breast"[All Fields] AND "neoplasms"[All Fields]) OR "breast neoplasms"[All Fields] OR ("breast"[All Fields] AND "cancer"[All Fields]) OR "breast cancer"[All Fields]) AND 2008[pdat]'],
Len = 6.


?- 
    date(Date), pubmed_search( prolog, Ids ), length( Ids, Len ), write( number_of:Len ), nl.

number_of:100
Date = date(2012, 7, 10),
Ids = ['22586414', '22462194', '22215819', '21980276', '21499053', '21353661', '20123506', '20123505', '19408879'|...],
Len = 100.

?- 
    date(Date), pubmed_search( prolog, Ids, retmax(200) ),
    length( Ids, Len ), write( number_of:Len ), nl.

number_of:120
Date = date(2012, 7, 10),
Ids = ['22586414', '22462194', '22215819', '21980276', '21499053', '21353661', '20123506', '20123505', '19408879'|...],
Len = 120.

==
*/
pubmed_search( Sterm, Ids, OptsIn ) :-
     to_list( OptsIn, Opts ),
     pubmed_search_defaults( Defs ),
     append( Opts, Defs, All ),
     eutils( Eutils ),
     ( ground(Sterm) -> true; type_error(ground,Sterm) ),
     search_term_to_query( Sterm, Query ),
     memberchk( retmax(Ret), All ), 
     atomic_list_concat( [Eutils,'esearch.fcgi?db=pubmed&retmax=',Ret,'&term=',Query], Url ),
     memberchk_optional( tmp_file(Tmp), All ),
     memberchk( verbose(Verb), All ),
     true_writes( Verb, Url ),
     get_url_in_tmp( Url, Verb, Tmp ),
     true_writes( Verb, tmp_file(Tmp) ),
     load_xml_file( Tmp, Xml ),
     ( (memberchk(qtranslation(QTrans),All),
        QT = 'QueryTranslation',
        search_element_in_list(Xml,QT,[],element(_,_,QTrans))) -> true; true
     ),
     all_subs_in_xml_single( Xml, 'IdList', 'Id', NastyIds ),
     flatten( NastyIds, Ids ),
     memberchk_optional( tmp_keep(Keep), All),
     true_atom_keeps_file( Keep, Tmp ).

/** pubmed_summary_display( +Ids ).

     Short for =|pubmed_summary_display( Ids, _Summary, [] ).|=
     
*/
pubmed_summary_display( Ids ) :-
     pubmed_summary_display( Ids, _Summary, [] ).

/** pubmed_summary_display( +Ids, -Summary ).

     Short for =|pubmed_summary_display( Ids, Summary, [] ).|=
*/
pubmed_summary_display( Ids, Summary ) :-
     pubmed_summary_display( Ids, Summary, [] ).

/** pubmed_summary_display( +Ids, -Summary, +Opts ).

     A wrapper around pubmed_summary_info/3. It call this predicate with same arguments before 
     displaying the Summary information. Opts can be a single term option or a list of such terms.
     In addition to pubmed_summary_info/3 options this wrapper also recognises the term : 

     * display(Disp) A list of article information keys that will displayed one on a line for each Id in =Ids=.

==
?-
     date(Date), pubmed_search((programming,'Prolog'), Ids), Ids = [A,B,C|_], pubmed_summary_display( [A,B,C] ).

----
0:22215819
[Evaluating bacterial gene-finding HMM structures as probabilistic logic programs.]
[Mørk S,Holmes I]
----
1:21980276
[War of ontology worlds: mathematics, computer code, or Esperanto?]
[Rzhetsky A,Evans JA]
----
2:15360781
[Medical expert systems developed in j.MD, a Java based expert system shell: application in clinical laboratories.]
[Van Hoof V,Wormek A,Schleutermann S,Schumacher T,Lothaire O,Trendelenburg C]
----
Date = date(2012, 7, 10),
Ids = ['22215819', '21980276', '15360781', '11809317', '9783213', '9293715', '9390313', '8996790', '15048396'|...],
A = '22215819',
B = '21980276',
C = '15360781'.

==

==
?-
     pubmed_cited_by( 20195494, These ), pubmed_summary_display( These, _, [display(['Title','Author','PubDate'])] ).
==

*/

pubmed_summary_display( Ids, Summary, OptsIn ) :-
     to_list( OptsIn, Opts ),
     pubmed_summary_display_defaults( Defs ),
     append( Opts, Defs, All ),
     memberchk( display(Disp), All ),
     pubmed_summary_info( Ids, Summary, Opts ),
     pubmed_summary_display_info( Summary, Disp ).

pubmed_summary_display_info( Summary, Disp ) :-
     write( '----' ), nl,
     nth0( N, Summary, Id-Rec ),
     write( N:Id ), nl,
     findall( _, (member(D,Disp),member(D-Val,Rec),write(Val),nl), _ ), 
     write( '----' ), nl, fail.
pubmed_summary_display_info( _Summary, _Disp ).


%% pubmed_cited_by( +Id, -Ids ).
%
% Redirects to pubmed_cited_by( Id, Ids, [] ).
%
pubmed_cited_by( Id, Ids) :-
     pubmed_cited_by( Id, Ids, [] ).

/** pubmed_cited_by( +Id, -Ids, +Options ). 

Ids is the list of pubmed ids that cite Id. 
Options is a term option or list of terms from the following;

     * verbose(Verb)  be verbose

==
?-
     date(D), pubmed_cited_by( 12075665, By ).

D = date(2012, 7, 9),
By = ['19497389'].
==
*/  
pubmed_cited_by( Id, Ids, OptsIn ) :-
     to_list( OptsIn, Opts ),
     ( memberchk(verbose(Verb),Opts) -> true; Verb = false ),
     eutils( Eutils ),
     Query = 'elink.fcgi?report=xml&mode=text&tool=curl&db=PubMed&cmd=neighbor&linkname=pubmed_pubmed_citedin&id=',
     atomic_list_concat( [Eutils,Query,Id], Url ),
     get_url_in_tmp( Url, Verb, Tmp ),
     load_xml_file( Tmp, Xml ),
     once( search_element_in_list( Xml, 'LinkSetDb', [], element(_,_,LXml) ) ),
     findall( CId, search_element_in_list(LXml,'Id',[],element(_,_,[CId])),Ids ),
     delete_file( Tmp ).

%% pubmed_cites( +Id, -Ids ).
%
% Redirects to pubmed_cites( Id, Ids, [] ).
%
pubmed_cites( Id, Ids ) :-
     pubmed_cites( Id, Ids, [] ).

%% pubmed_cites( +Id, -Ids ).
/** pubmed_cites( +Id, -Ids, +Options ). 

Ids is the list of pubmed Ids that are cited by Id. 
Options is a term option or list of terms from the following;

     * verbose(Verb)  be verbose

==
?-
     date(D), pubmed_cites( 20195494, Ids ), length( Ids, Len ), write( D:Len ), nl, fail.

date(2012,8,15):35
==

*/

pubmed_cites( Id, Ids, OptsIn ) :-
     to_list( OptsIn, Opts ),
     ( memberchk(verbose(Verb),Opts) -> true; Verb = false ),
     eutils( Eutils ),
     % Query = 'elink.fcgi?report=xml&mode=text&tool=curl&db=pmc&DbFrom=pubmed&Cmd=link&linkname=pubmed_pmc_refs&id=',
     Query = 'elink.fcgi?report=xml&mode=text&tool=curl&db=PubMed&Cmd=neighbor&linkname=pubmed_pubmed_refs&id=',
     atomic_list_concat( [Eutils,Query,Id], Url ),
     get_url_in_tmp( Url, Verb, Tmp ),
     load_xml_file( Tmp, Xml ),
     once( search_element_in_list( Xml, 'LinkSetDb', [], element(_,_,LXml) ) ),
     findall( CId, search_element_in_list(LXml,'Id',[],element(_,_,[CId])),Ids ),
     delete_file( Tmp ).

/** pubmed_summary_info( +Id, -Results, +Opts ).
 
   Results are the summary information for pubmed id Id.
   The predicate communicates with pubmed via the http interface with curl.
   Results are deposited in xml files which are subsequently parsed to 
   produce the termed =Results=. Id can also be a list of Ids in which case 
   the result is a list of Id-Results pairs.
  
   Options is a single term, or list of the following terms:
    
          * names(Names)  list of names to be found in the xml file. 
 
          * retmax(Retmax) the maximum number of records that will be returned def: 100

          * tmp_file(Tmp)  temporary file to be used for saving xml files. If Tmp is a variable, or option is missing, a temporary file is created with tmp_file_stream/3.
 
          * tmp_keep(Keep)  if true, keep the temporary xml file, otherwise, and by default, delete it.

          * verbose(Verb) When =true= be verbose.
 
==
?- 
  date(Date),  Opts = names(['Author','PmcRefCount','Title']),
  pubmed_summary_info( 12075665, Results, Opts ),
  write( date:Date ), nl, member( R, Results ), write( R ), nl, fail.

date:date(2012,7,9)
Author-[Kemp GJ,Angelopoulos N,Gray PM]
Title-[Architecture of a mediator for a bioinformatics database federation.]
Source-[IEEE Trans Inf Technol Biomed]
Pages-[116-22]
PubDate-[2002 Jun]
Volume-[6]
Issue-[2]
ISSN-[1089-7771]
PmcRefCount-[1]
PubType-[Journal Article]
FullJournalName-[IEEE transactions on information technology in biomedicine : a publication of the IEEE Engineering in Medicine and Biology Society]
false.

==
*/
pubmed_summary_info( Ids, Results, Opts ) :-
     % fixme use _defaults
     ( memberchk(names(Names),Opts) -> true; default_names(Names) ),
     ( memberchk(tmp_file(Tmp),Opts) -> true; true ),
     ( is_list(Ids) -> 
          findall( Id-Res, (member(Id,Ids),summary_info(Tmp,Id,Names,Res,Opts)), Results )
          ;
          summary_info( Tmp, Ids, Names, Results, Opts )
     ).

% Section: non-interface predicates...
summary_info( Tmp, Id, WhichIn, Results, Opts ) :-
     to_list( WhichIn, Which ),
     eutils( Eutils ),
     ( memberchk(retmax(RMax),Opts) -> true; RMax = 100 ),
     atomic_list_concat( ['esummary.fcgi?report=xml&mode=text&tool=wget&retmax=',RMax,'&db=PubMed&id='], Query ),
     atomic_list_concat( [Eutils,Query,Id], Url ),
     get_url_in_tmp( Url, false, Tmp ), % fixme
     load_xml_file( Tmp, Xml ),
     Elem = element(_,_,[Entry]),
     findall( Name-Info, ( member(Name,Which),
                           findall( Entry,
                             search_element_in_list(Xml, 'Item', ['Name'=Name], Elem ),
                                   Info )
                         ), Results ),
     ( memberchk(tmp_keep(true),Opts) -> true; delete_file(Tmp) ).

% Section: auxiliaries

%% search_term_to_query( Sterm, Query ).
%
%  Convert a pubmed search term to an atomic query that can be passed through http.
%
search_term_to_query( (A,B), Query ) :-
     !,
     search_term_to_query( A, Aq ),
     search_term_to_query( B, Bq ),
     atomic_list_concat( [Aq,'+AND+',Bq], Query ).
search_term_to_query( (A;B), Query ) :-
     !,
     search_term_to_query( A, Aq ),
     search_term_to_query( B, Bq ),
     atomic_list_concat( [Aq,'OR',Bq], '+', Query ).
search_term_to_query( (A=B), Query ) :-
     !,
     atomic_list_concat( [B,'\\[',A,'\\]'], Query ).
search_term_to_query( C, Query ) :-
     to_list( C, Clist ),
     atomic_list_concat( Clist, '+', Query ).

%% memberchk_optional( Elem, List ).
%
%  Unifies Elem with the first matching term in List if one exists.
%  The predicate always succeeds exactly once.
%
memberchk_optional( Elem, List ) :-
     memberchk( Elem, List ),
     !.
memberchk_optional( _Elem, _List ).
     
true_atom_keeps_file( Keep, _File ) :-
     Keep == true,
     !.
true_atom_keeps_file( _Keep, File ) :-
     delete_file( File ).

to_list( Either, List ) :-
	( (var(Either);(Either\=[_H|_T],Either\==[]) ) ->
		List = [Either]
		;
		List = Either
	).

true_writes( true, Report ) :-
     !,
     write( Report ), nl.
true_writes( _Opts, _Report ).

%% get_url_in_tmp(+URL, +Verbose, -In).
%
% From SWI-Prolog's doc files (July 2012). tmp_file_stream/3.
%
get_url_in_tmp(URL, Verb, File) :-
        ( var(File) ->
               tmp_file_stream(text, File, Stream),
               close(Stream)
               ;
               true
        ),
        ( Verb==true -> Args = ['-o',File,URL] ; Args = ['-s','-o',File,URL] ),
        true_writes( Verb, process_create(path(curl),Args,[]) ),
        process_create( path(curl), Args, [] ),
        exists_file( File ).

/*
http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=pubmed&term=science[journal]+AND+breast+cancer+AND+2008[pdat]
        
        */

all_subs_in_xml_single( Xml, Single, SubSel, Subs ) :-
     once( search_element_in_list( Xml, Single, [], element(_,_,Nest) ) ),
     findall( Sub, 
               search_element_in_list(Nest,SubSel,[],element(_,_,Sub)), 
              Subs ).

%% search_element_in_list(+Term, +Name, +Attrs, -Elem ).
% 
%  Find an element in an sgml file that have specific Attrs.
%  Got from hostip.pl .
%
search_element_in_list([Content|MoreContent], Name, ListAttributes, Element) :-
	(   search_element(Content, Name, ListAttributes, Element)
	;   search_element_in_list(MoreContent, Name, ListAttributes, Element)
	).

search_element(HTML, Name, ListAttributes, HTML) :-
	arg(1, HTML, Name),
	arg(2, HTML, HTML_Attributi),
	forall(member(Attribute, ListAttributes),
	       memberchk(Attribute, HTML_Attributi)).
search_element(HTML, Name, ListAttributes, Element) :-
	arg(3, HTML, Contents),
	search_element_in_list(Contents, Name, ListAttributes, Element).
